#include <parallel/algorithm>
#include <set>
#include "header.h"

/* Precomputation */
static vector < vector<DM_VERTEX> > left_list_fromIdentity_precomputed;

DM_VERTEX DM_transpose(DM_VERTEX V_input) {
    DM_VERTEX V_output;
    for (int col=0; col<fieldsize; col++) {
        MR temp_row=0;
        for (int row=0; row<fieldsize; row++) {
            if ((V_input.state[row]>>(fieldsize-1-col)) &0x1) {
                temp_row ^= 1;
            }
            if (row<fieldsize-1) temp_row = (temp_row<<1);
        }
        V_output.state[col] = temp_row;
    }

    return V_output;
}

DM_VERTEX XOR_rows(DM_VERTEX V_input, int R_i, int R_j) {
    DM_VERTEX V_output;
    if ((R_i>=fieldsize)||(R_j>=fieldsize)) {
        cout<<"XOR_rows: input is too large, no action is done."<<endl;
        return V_output;
    }
    if (R_i==R_j) {
        cout<<"XOR_rows: inputs are the same, no action is done."<<endl;
        return V_output;
    }
    for (int i=0; i<fieldsize; i++) {
        V_output.state[i] = V_input.state[i];
    }
    V_output.state[R_i]^=V_output.state[R_j];

    return V_output;
}

void sort_V_SortNetwork(DM_VERTEX& V) {
#define SWAP(x,y) {                  \
    if (V.state[y] < V.state[x]) {   \
        MR tmp = V.state[x];         \
        V.state[x] = V.state[y];     \
        V.state[y] = tmp;            \
    }                                \
}while(0);

#if fieldsize==8
    SWAP(1,0);SWAP(3,2);SWAP(5,4);SWAP(7,6);
    SWAP(2,0);SWAP(3,1);SWAP(6,4);SWAP(7,5);
    SWAP(2,1);SWAP(6,5);SWAP(4,0);SWAP(7,3);
    SWAP(5,1);SWAP(6,2);
    SWAP(4,1);SWAP(6,3);
    SWAP(4,2);SWAP(5,3);
    SWAP(4,3);
#elif fieldsize==4
    SWAP(0, 1);
    SWAP(2, 3);
    SWAP(0, 2);
    SWAP(1, 3);
    SWAP(1, 2);
#endif

}

bool CompareDM_VERTEX(DM_VERTEX V1, DM_VERTEX V2) {
    for (int i=0; i<fieldsize; i++) {
        if (V1.state[i] > V2.state[i]) return true;
        else if (V1.state[i] < V2.state[i]) return false;
    }
    return false;
}

bool isVertexIdentity(DM_VERTEX V_s) {
    for (int i=0; i<fieldsize;i++) {
        if(V_s.state[i] != ((1<<(fieldsize-1))>>i)) {
            return false;
        }
    }
    return true;
}

void insertDM_VERTEX(vector<DM_VERTEX> &L, DM_VERTEX V) {
    vector<DM_VERTEX>::iterator it = lower_bound(L.begin(), L.end(), V, CompareDM_VERTEX); // find proper position in descending order
    L.insert(it, V); // insert before iterator it
}

DM_VERTEX translateVertex(DM_VERTEX V, DM_VERTEX root) {

    int index;
    MR x;
    MR mask;
    DM_VERTEX tmp;
    tmp.parent = V.parent;
    tmp.XORpos = V.XORpos;
    tmp.outpos = V.outpos;

    for(int i=0; i<fieldsize; ++i) {

        x=V.state[i];

        mask=(1<<(fieldsize-1));
        index=0;
        tmp.state[i]=0;

        do {
            if(x&mask) tmp.state[i]^=root.state[index];
            mask>>=1;
            index++;
        } while(mask);
    }/*i*/

    sort_V_SortNetwork(tmp);
    return tmp;

}/* translateVertex */

DM_VERTEX translateVertex(DM_VERTEX V, MR translation[]) {

    DM_VERTEX tmp;
    for(int i=0; i<fieldsize; ++i) {
        tmp.state[i]=translation[V.state[i]];
    }/*i*/

    sort_V_SortNetwork(tmp);
    return tmp;

}/* translateVertex */


//grow the next level version3 (parallel grow, then reduce)
//sort and remove repetition at the end
//Remove elements appearing in any previous list
void next_level_list_fast(vector< vector<DM_VERTEX> > &Full_L, int print_details) {

    (void)print_details; /* remove warning */

    //add a empty vector list
    int stage = Full_L.size();  //eg. stage 3 =  having level 0,1 and 2
    Full_L.resize(stage+1); //now we want to grow level 3

    unsigned long long list_size = Full_L[stage-1].size();

    /* As many temporary containers as we use threads */
    vector<DM_VERTEX> tmp_storage[NUM_THREADS];

    #pragma omp parallel for shared(tmp_storage, list_size, Full_L, stage)
    for(unsigned long long DM_index=0; DM_index<list_size; DM_index++) {
        for (int outpos=0; outpos<fieldsize; outpos++) {

            const DM_VERTEX current_vertex = Full_L[stage-1][DM_index];

            for (int XORpos=0; XORpos<fieldsize;XORpos++) {
                if (XORpos==outpos) continue;

                DM_VERTEX V_temp = XOR_rows(current_vertex, outpos, XORpos);

                V_temp.parent = DM_index;
                V_temp.XORpos = XORpos;
                V_temp.outpos = outpos;
                sort_V_SortNetwork(V_temp);

                /* Store the result in the container for that thread */
                tmp_storage[omp_get_thread_num()].push_back(V_temp);
            }
        }
    }/* end parallel for */

    /* Here, we need to merge all the partial containers filled in all the threads */
    vector<DM_VERTEX> tmp;
    for(int tid=0; tid<NUM_THREADS; ++tid) {

        cout << "[-]     Merging partial result from Thread " << tid << "/" << (NUM_THREADS-1) << endl;

        /* Sort the current partial container */
        __gnu_parallel::sort(tmp_storage[tid].begin(), tmp_storage[tid].end(), CompareDM_VERTEX);

        /* Remove repeating element in that container */
        tmp_storage[tid].erase(
                unique(tmp_storage[tid].begin(), tmp_storage[tid].end(), EqualDM_VERTEX()),
                tmp_storage[tid].end()
                );

        /* Merge with final container (sorted and uniqued by construction) */
        __gnu_parallel::set_union(tmp_storage[tid].begin(), tmp_storage[tid].end(),
                                  Full_L[stage].begin(), Full_L[stage].end(),
                                  back_inserter(tmp),
                                  CompareDM_VERTEX);

        Full_L[stage].swap(tmp);
        vector<DM_VERTEX>().swap(tmp); /* swap trick to free */
        vector<DM_VERTEX>().swap(tmp_storage[tid]); /* swap trick to free */

    }/*tid*/

    for(int st=0; st<stage; ++st) {
        vector<DM_VERTEX> diff;
        __gnu_parallel::set_difference(
            Full_L[stage].begin(), Full_L[stage].end(),
            Full_L[st].begin(), Full_L[st].end(),
            inserter(diff, diff.begin()), /* Insert in diff the elements in Full_L[stage] that are NOT in Full_L[st] */
            CompareDM_VERTEX
        );

        Full_L[stage]=diff;
        vector<DM_VERTEX>().swap(diff); /* Free memory */

    }/*st*/

    // printf("Full_L[%d].size() = %ld\n", stage, Full_L[stage].size());

    return;
}

/*
** return:
**         -1 if V1<V2
**          0 if V1==V2
**          1 if V1>V2
*/
int matchDM_VERTEX(DM_VERTEX V1, DM_VERTEX V2) {
    for (int i=0; i<fieldsize; i++) {
        if (V1.state[i] > V2.state[i]) return 1;
        else if (V1.state[i] < V2.state[i]) return -1;
    }
    return 0;
}

/*
** Match the two latests vectors
** Note: the vectors are stored in decreasing order
*/
void match_fast(vector < vector<DM_VERTEX> > L_list,
                int L_level,
                DM_VERTEX V_left,
                vector < vector<DM_VERTEX> > R_list,
                int R_level,
                DM_VERTEX V_right,
                vector<DM_VERTEX>& Stage_state,
                int print_details) {

    int L_pointer=0;
    int R_pointer=0;

    int L_size = L_list[L_level].size();
    int R_size = R_list[R_level].size();

    int match_val = 1;

    while ((L_pointer<L_size) && (R_pointer<R_size) && (match_val!=0)) {

        match_val = matchDM_VERTEX(L_list[L_level][L_pointer], R_list[R_level][R_pointer]);

        if(match_val==1)       L_pointer++; //L is larger
        else if(match_val==-1) R_pointer++; //R is larger

    } /* while no match found */

    if (match_val==0) {

        if (print_details>0) cout << "Found match! L_level " << L_level << " , R_level " << R_level << endl;

        DM_VERTEX V_temp;

        /****************************************************/
        V_temp = L_list[L_level][L_pointer];

        if(R_level>0) Stage_state.insert(Stage_state.begin(), V_temp);
        if(print_details) {printf("Match in L_list (parent=%7d):    ", V_temp.parent); printVertex(V_temp, 0);}

        while(L_level>0) {

            /* Get pointer on the parent of the current node */
            L_pointer = V_temp.parent;
            L_level--;
            if(L_level==0) break;

            /* Get the actual parent node */
            V_temp = translateVertex(left_list_fromIdentity_precomputed[L_level][L_pointer], V_left);

            /* Push it to the found sequence of nodes */
            Stage_state.insert(Stage_state.begin(), V_temp);

        } /* while L_level */

        /****************************************************/
        V_temp = R_list[R_level][R_pointer];
        if(print_details) {printf("Match in R_list (parent=%7d):   ", V_temp.parent); printVertex(V_temp, 0);}

        while(R_level>0) {

            /* Get pointer on the parent of the current node */
            R_pointer = V_temp.parent;
            R_level--;
            if(R_level==0) break;

            /* Get the actual parent node */
            V_temp = translateVertex(left_list_fromIdentity_precomputed[R_level][R_pointer], V_right);

            /* Push it to the found sequence of nodes */
            Stage_state.push_back(V_temp);

        } /* while R_level */

        /****************************************************/
        //add starting and ending state to the Stage_state
        Stage_state.insert(Stage_state.begin(), L_list[0][0]);
        Stage_state.push_back(R_list[0][0]);

        // printf("-- In match: --------------------\n");
        // for(auto v: Stage_state) printVertex(v, 0);
        // printf("---------------------------------\n");

    } else {

        /* No match */

        if (print_details>0) {
            cout << "L_level " << L_level << ", R_level " << R_level << " has no match" << endl;
        }

    } /* if match */

    return;

} /* match_fast */


/*
** Precompute MAX_STEP level of the graph rooted on the Identity.
** Since this tool only considers LINEAR layers, we can translate this graph
** by a linear operation for *any* root.
*/
void MITM_precompute(int MAX_STEP, int print_details) {

    DM_VERTEX V_s;
    for (int i=0; i<fieldsize;i++) {
        V_s.state[i]=(1<<(fieldsize-i-1));
    }

    left_list_fromIdentity_precomputed.resize(1);
    left_list_fromIdentity_precomputed[0].push_back(V_s);

    for(int step=1; step<=MAX_STEP; step+=2) {
        printf("[-] Precomputing Level %d...\n", step/2+1);
        next_level_list_fast(left_list_fromIdentity_precomputed, print_details);
    }

} /* MITM_precompute */

/*
** For a given root node (V_r) and a given level (i), this function
** returns the sorted vector of all the nodes reachable after i
** instructions from the root node V_r
**
** Prerequisite: calling the precomputation function MIMT_precompute()
*/
vector<DM_VERTEX> getListFromRoot(DM_VERTEX root, int level) {

    /* If the root is the Identity, then we already computed the list */
    if(isVertexIdentity(root)) {
        return left_list_fromIdentity_precomputed[level];
    }

    vector<DM_VERTEX> L;
    MR translation[1<<fieldsize];
    const vector<DM_VERTEX> precomputed = left_list_fromIdentity_precomputed[level];

    for(size_t x=0; x<(1<<fieldsize); ++x) {

        MR val=0;
        MR mask=(1<<(fieldsize-1));
        int index=0;

        do {
            if(x&mask) val^=root.state[index];
            mask>>=1;
            index++;
        } while(mask);

        translation[x]=val;

    }/*x*/

    /* Populate L form the translated values of precomputed */
    for(size_t x=0; x<precomputed.size(); ++x) {

        DM_VERTEX tmp=translateVertex(precomputed[x], translation);
        tmp.parent = precomputed[x].parent;
        tmp.XORpos = precomputed[x].XORpos;
        tmp.outpos = precomputed[x].outpos;

        L.push_back(tmp);

    }/*x*/

    /* Sort the current container */
    __gnu_parallel::sort(L.begin(), L.end(), CompareDM_VERTEX);

    return L;
}


void MITM_search_fast(DM_VERTEX V_s,
                      DM_VERTEX V_e,
                      vector<DM_VERTEX>& Stage_state,
                      int MAX_STEP,
                      int print_details) {


    //sort the input vertices
    sort_V_SortNetwork(V_s);
    sort_V_SortNetwork(V_e);

    if (print_details>0) {
        cout<<"Start vertex (after sorting): "; printVertex(V_s, 0);
        cout<<"End vertex (after sorting):   "; printVertex(V_e, 0);
    }

    vector< vector<DM_VERTEX> > left_list;
    vector< vector<DM_VERTEX> > right_list;

    left_list.resize(1);
    right_list.resize(1);

    left_list[0].push_back(V_s);
    right_list[0].push_back(V_e);

    int step=1;
    while ((Stage_state.size()==0) && (step<=MAX_STEP)) {

        //assume V_s and V_e are different
        if ((step%2) == 1) {

            /* Get the new left list */
            size_t s=left_list.size();
            left_list.resize(s+1);
            left_list[s]=getListFromRoot(V_s, s);

        } else {

            /* Get the new right list */
            size_t s=right_list.size();
            right_list.resize(s+1);
            right_list[s]=getListFromRoot(V_e, s);
        }

        match_fast(left_list, left_list.size()-1, V_s,
                   right_list, right_list.size()-1, V_e,
                   Stage_state,
                   print_details);

        step++;
    }

    if ((Stage_state.size())==0) {
        if (print_details>0) {
            cout<<"optimal number of steps > "<<MAX_STEP<<endl;
        }
    }

    return ;

} /* MITM_search_fast */


void MITM_search_shell(DM_VERTEX V_s,
                       DM_VERTEX V_e,
                       vector<DM_VERTEX>& Stage_state,
                       int MAX_STEP,
                       int print_details) {

    MITM_search_fast(V_s, V_e, Stage_state, MAX_STEP, print_details);

}
